from typing import Optional

import requests


def _request(
    method: str,
    url: str,
    headers=None,
    input: str = "json",
    data: dict = None,
    session: Optional[requests.Session] = None
):
    res = (session or requests).request(
        method,
        url=url,
        json=data if input == "json" else None,
        data=data if input == "form" else None,
        headers=headers,
    )
    return res


def post(
    url: str,
    headers=None,
    input: str = "json",
    output="json",
    data: dict = None,
    session: Optional[requests.Session] = None
):
    res = _request(
        "POST", url,
        headers, input,
        data, session,
    )
    return _handle_res(res, output=output)


def patch(
    url: str,
    headers=None,
    input: str = "json",
    output="json",
    data: dict = None,
    session: Optional[requests.Session] = None
):
    res = _request(
        "PATCH", url,
        headers, input,
        data, session,
    )
    return _handle_res(res, output=output)


def delete(
    url: str,
    headers=None,
    input: str = "json",
    output="json",
    data: dict = None,
    session: Optional[requests.Session] = None
):
    res = _request(
        "DELETE", url,
        headers, input,
        data, session,
    )
    return _handle_res(res, output=output)


def put(
    url: str,
    headers=None,
    input: str = "json",
    output="json",
    data: dict = None,
    session: Optional[requests.Session] = None
):
    res = _request(
        "PUT", url,
        headers, input,
        data, session,
    )
    return _handle_res(res, output=output)


def _handle_res(resp: requests.Response, output):
    if output == "response":
        return resp
    elif output == "none":
        return None
    if resp.ok:
        if output == "json":
            return resp.json()
        else:
            return resp.text


def get(
    url: str,
    output="json",
    headers=None,
    session: Optional[requests.Session] = None
):
    res = (session or requests).get(url, headers=headers)
    return _handle_res(res, output=output)


def get_ping(
    url: str,
    headers=None,
    session: Optional[requests.Session] = None
):
    res = (session or requests).get(url, headers=headers)
    return res.ok



import logging
import time
from typing import List, Optional

from .base_functions import functions
from .context import Context
from .datastruct import VARIABLE_TOKEN, Recipe, Result, Step, Test

# TODO: move to async to handle more load
# TODO: implem dependency detection


class Executor:
    def __init__(self) -> None:
        self.context = Context()
        self.logger = logging.getLogger()

    def _prepare_args(self, args: dict, current=None) -> dict:
        res = current or {}
        # trouver ce qui est entre {{}} et remplacer comme un format
        # trouver ce qui est entre {} et mettre la variable à la place directement
        for k, v in args.items():
            if isinstance(v, str):
                if v.startswith(VARIABLE_TOKEN) and len(v) > 2:
                    res[k] = self.context.get(v[1:])
                else:
                    res[k] = v.format_map(self.context.data)
            elif isinstance(v, dict):
                res[k] = self._prepare_args(v)
            else:
                res[k] = v
        return res

    def _do_test(self, step: Step, test: Test) -> bool:
        result = False
        try :
            func = functions[test.function]
            args = self._prepare_args(test.with_)
            result = func(**args)
        except:
            # TODO: log
            pass
        return result

    def __configure(self, args: Optional[dict]) -> None:
        if args is not None:
            for k, v in args.items():
                self.context.put(k, v)

    def do_step(self, step: Step) -> Result:
        start = time.time()
        func = functions[step.function]
        args = self._prepare_args(step.with_)
        self.logger.debug(f'{args}')
        result = func(**args)
        self.logger.debug(f'{step.name} did return {result}')

        if step.returns is not None:
            self.context.put(step.returns, result)

        test_passed = True
        if step.tests is not None:
            for test in step.tests:
                test_passed = self._do_test(step, test)
        
        end = time.time()
        return Result(
            name=step.name,
            success=test_passed,
            elapsed=end-start
        )

    def add_env(self, args: dict):
        self.__configure(args)

    def bind_args(self, args: dict):
        self.__configure(args)

    def execute_recipe(self, recipe: Recipe) -> List[Result]:
        results = []
        self.__configure(recipe.env)
        for step in recipe.steps:
            results.append(self.do_step(step))

        return results



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

VARIABLE_TOKEN = "$"

FunctionParams = Dict[str, Any]


class Test(BaseModel):
    function: str
    with_: FunctionParams

    class Config:
        fields = {
            'with_': 'with'
        }

class Step(BaseModel):
    name: str
    function: str
    with_: FunctionParams
    returns: Optional[str]
    tests: Optional[List[Test]]


    class Config:
        # rename field with to with_
        fields = {
            'with_': 'with'
        }


class Recipe(BaseModel):
    name: str
    kind: str
    # should create class for later
    trigger: dict
    steps: List[Step]
    args: List[str] = []
    env: Dict[str, Any] = {}

@dataclass
class Result:
    elapsed: float
    success: bool
    name: str

    def to_json(self):
        return {
            "name": self.name,
            "elapsed": self.elapsed,
            "success": self.success,
        }




from typing import Any


class Context:
    def __init__(self) -> None:
        self.data = {}

    def put(self, name: str, value: Any):
        self.data[name] = value

    # handle nested values with "." notation parent.nested
    def get(self, name: str):
        return self.data[name]



import unittest

import yaml

from .. import executor, utils

data = """
name: check site alive
kind: test.availability

trigger:
  every: 5m

args: []

steps:
  - name: http ping
    function: http.ping_get
    with:
      url: https://google.com
"""


http_lib_data = """
name: check site alive
kind: test.availability

trigger:
  every: 5m

args: []

steps:

  - name: test post
    function: http.post
    with:
      url: https://example.com
      output: text
    


#   - name: test put
#     function: http.put
#     with:
#       url: https://example.com
#       output: text
  
  - name: test patch
    function: http.patch
    with:
      url: https://example.com
      output: response
    returns: res

  - name: pick resp
    function: pick
    with:
      input: $res
      key: status_code
    returns: status_code
    tests:
      - function: not_equal
        with:
          a: $status_code
          b: 200
      - function: equal
        with:
          a: $status_code
          b: 405

#   - name: test delete
#     function: http.delete
#     with:
#       url: https://example.com
#       output: text
    """

class ExecutorTests(unittest.TestCase):
    def setUp(self):
        self.recipe = self._load_base_recipe()
        return super().setUp()

    def _load_base_recipe(self):
        return utils.load_recipe(yaml.safe_load(data))

    def test_load_recipe(self):
        recipe = utils.load_recipe(yaml.safe_load(data))
        self.assertEqual(recipe.name, 'check site alive')
        self.assertEqual(recipe.kind, 'test.availability')
        self.assertEqual(recipe.args, [])
        self.assertDictContainsSubset(recipe.steps[0].with_, {"url": "https://google.com"})

    def test_basic_recipe_executor(self):
        e = executor.Executor()
        res = e.execute_recipe(self.recipe)
        self.assertTrue(res[0].success)

    def test_http_functions(self):
        recipe = utils.load_recipe(yaml.safe_load(http_lib_data))
        e = executor.Executor()
        res = e.execute_recipe(recipe)
        print(res)
        self.assertTrue(res[0].success)
        # here as we requested
        # self.assertFalse(res[1].success)
        self.assertTrue(res[2].success)



from typing import Any


def contains(input: str, inside: str) -> bool:
    return inside in input


def pick(input: dict, key: str) -> Any:
    return input.__getattribute__(key)


def equal(a: Any, b: Any):
    return a == b

def not_equal(a: Any, b: Any):
    return a != b




from requests.api import delete

from .base_functions import contains, equal, not_equal, pick
from .http_funcs import delete, get, get_ping, patch, post, put

functions = {
    "http.get": get,
    "http.post": post,
    "http.put": put,
    "http.delete": delete,
    "http.patch": patch,
    "http.ping_get": get_ping,

    # basic utils
    "pick": pick,
    "contains": contains,
    "equal": equal,
    "not_equal": not_equal,
}
